// mmap을 사용하고자 할 때 선언
#define USE_MMAP

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef WIN32
  #undef USE_MMAP
#else
  #include <unistd.h> // close
  #include <sys/stat.h> // open
  #include <fcntl.h> // open
#endif

#ifdef USE_MMAP
  #include <sys/mman.h> // mmap
#endif

#include "bin2txt.h"

typedef struct {
  long mmapsize;
  int str_count;
  int *str_index;
  char *contents;
} B2TtFILE;

/////////////////////////////////////////////////////////////////////
// 파일명이 유효한지를 검사
// [리턴] 성공(1), 오류값
static int __IsValidFileName(const char *filename)
{
  if (filename == NULL || filename[0] == 0)
    return B2TeNULL_DICNAME;
  else return 1;
}

/////////////////////////////////////////////////////////////////////
#ifndef USE_MMAP
// #include <stdio.h>
// 파일 크기 반환
// [리턴] 파일 크기 (-1: 에러)
static long __GetFileSize
( 
  const char *filename // [입력] 파일 명
) 
{
  long filesize = 0;
  FILE *fp;
  
  if ((fp = fopen( filename, "rb")) == NULL) 
  {
    if (fp == NULL) return -1;
  }
  
  // 화일의 크기를 알아냄
  fseek( fp, 0, SEEK_END); // 화일의 끝
  filesize = ftell( fp);   // 화일의 위치
  fseek( fp, 0, SEEK_SET); // 화일의 처음
  
  fclose( fp);
  
  return filesize;
}
#endif

/////////////////////////////////////////////////////////////////////
// [리턴] 핸들 번호 or error code
void *b2t_Open
(
  const char *Filename // [입력] 파일 명
)
{
  char *contents;
  int str_count;
  int *str_index;
  long filesize; 
  
  B2TtFILE *stream;

#ifdef USE_MMAP
  int fileHandle = -1;
  struct stat statbuf;
#else
  FILE *infp;
#endif
 
  if ( ! __IsValidFileName( Filename))
    return NULL;
  
  // 실제 리소스를 오픈한다. ////////////////////////////////////////
#ifdef USE_MMAP
  // 파일 열기
  fileHandle = open( Filename, O_RDONLY);

  if ( fstat( fileHandle, &statbuf) < 0)
  {
    /**/fprintf( stderr, "Error: %d line in [%s]\n",__LINE__,__FILE__);
    
    return NULL;
  }
    
  filesize = statbuf.st_size; // 파일 사이즈
  
  if (filesize < 0) return NULL;
    
  // mmap
  contents = (char *) mmap( NULL, filesize, PROT_READ, MAP_SHARED, fileHandle, 0);

  if ( (void *) contents == MAP_FAILED)
  {
    /**/fprintf( stderr, "Error: %d line in [%s]; %s\n",__LINE__,__FILE__, strerror(errno));
    return NULL;
  }
  
  ///**/fprintf( stderr, "###### mmap position = %p\n", contents);
    
  close( fileHandle); // 파일 닫기
  
#else
  // 화일의 크기를 알아냄
  filesize = __GetFileSize( Filename);
  if (filesize == -1) return NULL;
      
  // 바이너리 파일 읽기
  infp = fopen( Filename, "rb"); /* 화일 열기 */
  if (infp == NULL) return NULL;
    
  contents = (char *) malloc( filesize+1); // 메모리 할당
  if (contents == NULL)
  {
    /**/fprintf( stderr, "Error: %d line in [%s]\n",__LINE__,__FILE__);
    return NULL;
  }

  fread( contents, filesize, 1, infp); // 화일 전체를 읽어들임
  contents[filesize] = 0; // NULL
  fclose( infp); // 화일 닫기
#endif
  
  str_index = (int *) contents;
  str_count = *str_index; // 엔트리 수
  
  if (str_count <= 0) return NULL;

  str_index++;
  if (*str_index <= 0) return NULL;
  
  // 핸들 구조체에 저장
  stream = (B2TtFILE *) malloc( sizeof( B2TtFILE)); // memory allocation
  if (stream == NULL)
  {
    /**/fprintf( stderr, "Error: %d line in [%s]\n",__LINE__,__FILE__);
    return NULL;
  }
  
  stream->contents = contents;
  stream->mmapsize = filesize;
  stream->str_count = str_count;
  stream->str_index = str_index;
  
  return (void *)stream;
}

////////////////////////////////////////////////////////////////////////////
// [출력] 문자열의 수
int b2t_GetStringCount( void *Rsc)
{
  return ((B2TtFILE *)Rsc)->str_count;
}

////////////////////////////////////////////////////////////////////////////
// [리턴] 문자열
// n번째 문자열을 리턴
char *b2t_GetString
(
  void	*Rsc, 
  int	N            // [입력] 문자열 인덱스
)
{
  B2TtFILE *rsc_ptr = (B2TtFILE *) Rsc;
  
  // 최대 수보다 큰 인덱스를 가리키면
  if (rsc_ptr->str_count <= N) return NULL;
  
  return &(rsc_ptr->contents[rsc_ptr->str_index[N]]);
}

////////////////////////////////////////////////////////////////////////////
// [리턴] 성공 : 1, 오류값
int b2t_Close( void *Rsc)
{
  B2TtFILE *rsc_ptr = (B2TtFILE *) Rsc;
  char *contents;

  contents = rsc_ptr->contents;
  
#ifdef USE_MMAP
  int rc = munmap( contents, rsc_ptr->mmapsize);
  if (rc == -1) return B2TeMUNMAP_FAILED;
#else
  // 메모리 해제
  if ( contents) free( contents); // 이거 하나로 해결
  contents = NULL;
#endif

  rsc_ptr->mmapsize = -1;
  rsc_ptr->str_count = -1;
  rsc_ptr->str_index = NULL;
  
  free( rsc_ptr);
  rsc_ptr = NULL;

  return 1;
}
