#ifndef __DAFST__
#define __DAFST__

#ifdef __cplusplus
extern "C" {
#endif

#define FSTcNOT_EXIST (-1) // 존재하지 않는 키에 대한 해쉬값

// read/write 모드로 새로운 FST 만들기
// 아래의 함수들 인자중 pTransducer는 fst_New()의 반환값임
extern void *fst_New( void);

// FST에 할당된 메모리를 free
extern void fst_Close( void *pTransducer);

// FST에 새로운 키 삽입하기
// key : 삽입할 키
// return value : 1 = 성공, 0 = 실패
extern int fst_InsertKey( void *pTransducer, const char *key);

// FST에서 키를 제거하기
// Key : 삭제할 키
// Key가 없으면 아무런 작업도 하지 않음(반환값은 true)
// Hash : 삭제할 키의 해쉬값
//   -> 중복된 키가 여러 개일 경우 Hash값에 해당하는 키만 삭제
// Hash가 FSTcNOT_EXIST이거나 Key의 해쉬값이 아니면 중복된 키를 모두 삭제
// 삭제 도중 오류가 발생하면 false를 반환
extern int fst_DeleteKey( void *pTransducer, const char *Key, int Hash);


// FST를 로딩하기
// ContentFilename : FST 파일 이름
// ContentFilename에 해당하는 파일이 없으면 NULL 반환
// HashFilename : 부가 정보를 저장하는 파일 이름
// HashFilename이 NULL이면 read only 모드로 로딩
// HashFilename이 NULL이 아니면 read/write 모드로 로딩
// 반환값 : 성공 여부
extern void *fst_Open( const char *ContentFilename, const char *HashFilename);

// 로딩된 FST를 파일로 저장하기
// ContentFilename : FST 파일 이름
// HashFilename : 부가 정보를 저장하는 파일 이름
// FST가 read only 모드로 로딩되었거나 
// HashFilename이 NULL이면 부가 정보는 저장하지 않음
// 반환값 : 성공 여부
extern int fst_Save( void *pTransducer, const char *ContentFilename, 
		     const char *HashFilename);

// FST에 들어있는 엔트리의 개수
extern int fst_GetNumberOfEntry( void *pTransducer);

// FST에 있는 모든 키에 대해 순차적으로 Callback을 호출하는 함수
// pParam : Callback을 호출할 때 전달될 첫번째 인자
// Callback : 각각의 키에 대해 호출되는 callback 함수
//            s : 키
//            nItem : s에 해당하는 키의 개수
extern void fst_Traverse( void *pTransducer, void *pParam,
			  void (*Callback)(void *pParam, const char *s, int Hash, int nItem));

// FST가 올바른지 검사하는 함수
// 반환값 : FST가 올바른지 여부
extern int fst_Check( void *pTransducer);

/* 문자열 -> 해쉬값 */
/* *nItem : String에 일치하는 엔트리의 개수, nItem이 NULL이면 무시 */
/* 반환값 : String에 일치하는 첫번째 엔트리의 Hash Value 또는 FSTcNOT_EXIST(검색 실패) */
extern int fst_String2Hash( void *pTransducer, const char *String, int *nItem);

/* 해쉬값 -> 문자열 */
/* 반환값 : 해쉬값이 부적절할 경우 NULL, 그렇지 않으면 String을 반환 */
extern char *fst_Hash2String( void *pTransducer, int HashValue, char *String);

/* 패턴에 일치하는 키 탐색 */
/* Pattern : 검색할 엔트리의 패턴(와일드카드 *, ? 사용 가능) */
/* pParam : Callback을 호출할 때 첫번째 인자로 전달될 값 */
/* Callback : Pattern에 부합되는 엔트리가 탐색될 때마다 
              그 엔트리의 스트링과 해쉬값을 가지고 호출되는 callback */
/* 반환값 : pattern에 일치되는 엔트리의 개수 */
extern int fst_Pattern2Hash( void *pTransducer, const char *Pattern, void *pParam,
		             int (*Callback)(void *pParam, const char *s, int Hash));

/* tabular parsing을 위한 테이블 구축 */
/* fst에 등록된 키들을 기반으로 String을 tabular parsing하기 위한 table을 구축 */
/* String : tabular parsing을 수행할 문자열 */
/* pParam : Callback을 호출할 때 첫번째 인자로 전달될 값 */
/* Callback : String[From]~String[From+Length-1]에 해당하는 부분 문자열이 
              FST에 등록되어 있으면
              그 부분 문자열의 해쉬값을 Hash에 대입하여 호출되는 callback */
/*            Size : String의 길이 */
extern void fst_String2Tabular( void *pTransducer, const char *String, void *pParam, 
		                int (*Callback)(void *pParam, int Size, int From, int Length, int Value));

extern void fst_String2Tabular_uhc( void *pTransducer, const char *String, void *pParam, 
                                    int (*Callback)(void *pParam, int Size, int From, int Length, int Hash));

/* 문자열 -> 최장 일치 키의 해쉬값 */
/* nItem : String에 최장 일치하는 엔트리의 개수 */
/* 반환값 : String에 최장 일치하는 첫번째 엔트리의 Hash Value 또는 NULL_INDEX(검색 실패) */
/* 주의 : Last가 없는 버전 */
extern int String2LongestMatchedHash(void *pTransducer, const char *String, int *nItem);

/* 1차원 배열로 구성된 Table의 cell에 접근하는 매크로 
   From은 0부터 시작, Length는 1부터 시작 */
#define FSTmGetCELL(Table, Size, From, Length) ((Table)[(From)*(2*(Size)-(From)+1)/2+(Length)-1])

// 삼각테이블의 모양이 다르다.
//   1 2 3
// 0 * * *
// 1 * *
// 2 *
/* Size는 불변(전체 길이), From = 시작위치, Length = 문자열 길이 */
#define FSTmGetTabPos1(Size, From, Length)   ((From)*(2*(Size)-(From)+1)/2+(Length)-1)   

//   1 2 3
// 0 * * *
// 1   * *
// 2     *
/* raw = 행번호, column = 열번호, n = 전체 문자열 길이 */
#define FSTmGetTabPos2(row, column, n)  ((row)*(2*(n)-(row)-1)/2+(column)-1) 

#define FSTmGetTabNum(x) ((x)*((x)+1)/2)

// Key파일로부터 FST파일과 hash파일 생성
// return value : 1 for success, 0 for failure
extern int fst_Build
(
  char *KeyFileName, 
  char *FstFileName, 
  char *HashFileName
);

// 주어진 문자열 목록으로부터 FST를 구축한다.
// 반환 : NULL (오류), FST 포인터 (성공)
extern void *fst_Compile
(
  const char  **TermList,    // [input] 문자열 목록
  int           TermCnt      // [input] 문자열의 수
);

#ifdef __cplusplus
}
#endif // __cplusplus

#endif
